package uk.co.brmartin


import com.intellij.codeInsight.inline.completion.InlineCompletionEvent
import com.intellij.codeInsight.inline.completion.InlineCompletionProvider
import com.intellij.codeInsight.inline.completion.InlineCompletionProviderID
import com.intellij.codeInsight.inline.completion.InlineCompletionRequest
import com.intellij.codeInsight.inline.completion.elements.InlineCompletionElement
import com.intellij.codeInsight.inline.completion.elements.InlineCompletionGrayTextElement
import com.intellij.codeInsight.inline.completion.suggestion.InlineCompletionSuggestion
import com.intellij.codeInsight.inline.completion.suggestion.invoke
import com.intellij.openapi.util.TextRange
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.emitAll
import kotlinx.coroutines.flow.flow
import kotlinx.serialization.encodeToString
import okhttp3.Request.Builder
import okhttp3.RequestBody.Companion.toRequestBody
import uk.co.brmartin.ollama.RequestBody
import uk.co.brmartin.ollama.ResponseBody

class CompletionProvider : InlineCompletionProvider {

    override val id: InlineCompletionProviderID = InlineCompletionProviderID("Wingman")

    override suspend fun getSuggestion(request: InlineCompletionRequest): InlineCompletionSuggestion =
        InlineCompletionSuggestion {
            variant {
                emitAll(doFIM(request))
            }
        }

    override fun isEnabled(event: InlineCompletionEvent): Boolean = true

    private fun doFIM(request: InlineCompletionRequest): Flow<InlineCompletionElement> = flow {
        val prefixCode = request.document.getText(TextRange(0, request.startOffset))
        val suffixCode = request.document.getText(TextRange(request.endOffset, request.document.textLength))

        val request = requestFactory(prefixCode, suffixCode)

        request
            .let(HTTP_CLIENT::newCall)
            .execute()
            .use { response ->
                response
                    .body!!
                    .byteStream()
                    .reader()
                    .useLines { lines ->
                        for (line in lines) {
                            emit(line
                                .let { JSON.decodeFromString<ResponseBody>(it) }
                                .response
                                .let(::InlineCompletionGrayTextElement))
                        }
                    }
            }
    }

    private fun requestFactory(prefixCode: String, suffixCode: String) = Builder()
        .url(URL)
        .post(requestBodyFactory(prefixCode, suffixCode))
        .build()

    private fun requestBodyFactory(prefixCode: String, suffixCode: String) = JSON
        .encodeToString(RequestBody(MODEL, prefixCode, suffixCode, options))
        .toRequestBody(APPLICATION_JSON)
}

